from ptsa.data.timeseries import TimeSeries

try:
    from mne.io import read_raw_edf
except ImportError:
    read_raw_edf = None
# Turn off stupid mne RunTimeWarnings for EDF+
import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)
from datetime import datetime, date
from pyedflib import EdfWriter
import xarray as xr
import numpy as np
from IPython.display import display
import pyedflib

__all__ = ['TimeSeriesLF']

class TimeSeriesLF(TimeSeries):
    """A thin wrapper around :class:`xr.DataArray` for dealing with time series
    data and timeseries.
    Note that xarray internals prevent us from overriding the constructor which
    leads to some awkwardness: you must pass coords as a dict with a
    ``samplerate`` entry.
    Parameters
    ----------
    data : array-like
        Time series data
    coords : dict-like
        Coordinate arrays. This must contain at least a ``samplerate``
        coordinate.
    dims : array-like
        Dimension labels
    name : str
        Name of the time series
    attrs : dict
        Dictionary of arbitrary metadata
    encoding : dict
    fastpath : bool
        Not used, but required when subclassing :class:`xr.DataArray`.
    Raises
    ------
    AssertionError
        When ``samplerate`` is not present in ``coords``.
    See also
    --------
    xr.DataArray : Base class
    """

    def __init__(self, data, coords, dims=None, name=None,
                 attrs=None, encoding=None, fastpath=False):
        assert 'samplerate' in coords
        super(TimeSeriesLF, self).__init__(data=data, coords=coords, dims=dims,
                                           name=name, attrs=attrs, encoding=encoding,
                                           fastpath=fastpath)

    """Start Logan Monkey Patching"""

    # TODO: Change to automatic correct_channels as array
    @classmethod
    def from_edf(cls, filename, correct_channels=None, *args, **kwargs):
        """Takes a file located at filename and returns it as a TimeSeriesX object

        FIXME: automate the generation of 'events' based upon raw signal for _sleep data

        ------
        INPUTS
        filename: str, path to the edf file, e.g. '/Volumes/rhino/home2/loganf/data.edf'

        ------
        OUTPUTS
        ts: TimeSeriesX

        Parameters
        ----------
        filename : str

        """
        if read_raw_edf is None:
            raise RuntimeError("You must install mne to load from EDF")
        print('Extracting edf signal from {}'.format(filename))
        # Use MNE to read in raw edf file
        raw_data = read_raw_edf(filename, preload=False, verbose='ERROR', stim_channel=None,
                                *args, **kwargs)

        # Handling for clinical copying bizzare filenames
        if correct_channels is not None:
            matched = cls.get_eeg_ch_matches(raw_data.ch_names, correct_channels)
            all_chs = np.arange(len(raw_data.ch_names))
            bad_indx = [x for x in all_chs if x not in matched]
            exclude = (list(np.array(raw_data.ch_names)[bad_indx]))
            raw_data = read_raw_edf(filename, stim_channel=None, exclude=exclude, preload=True,
                                    verbose='ERROR', *args, **kwargs)

        # Handling for normal things
        elif correct_channels is None:
            # Mne defaults an empty channel for stim or annotations this will make sure it's removed
            # By default it's assigned to the last channel....
            if raw_data.ch_names[-1] in ('EDF Annotations', 'STI 014'):
                raw_data = read_raw_edf(filename, stim_channel=None, exclude=raw_data.ch_names[-1], preload=True,
                                        verbose='ERROR', *args, **kwargs)
            else:
                raw_data.load_data()

        # Get data, channel labels, sampling freq
        data, chs, sfreq = raw_data._data, raw_data.ch_names, np.round(raw_data.info['sfreq'])
        # Construct time from number of points and sampling freq
        time = np.arange(0, data.shape[-1]) / sfreq
        # Set coords for timeseries
        coords = {'channels': chs, "time": time, "samplerate": sfreq}
        # Create TimeSeriesX object
        ts = cls.create(data, sfreq, coords=coords,
                        dims=['channels', 'time'])
        print('Ready.')
        return ts

    @classmethod
    def concat(cls, ts_list, dim='events', *args, **kwargs):
        """Concatenates a list of TimeSeriesX objects along a dimension

        FIX ME: Change to check if any dim is np.recarray and reset any that are.
        -----
        INPUTS
        ts_list: list, a list of time_series seperated, e.g. list of sessions
        dim: str, dimension to concatenate over, you can also choose a new name (e.g. 'subjects'
             across all a list of all subjects). By default tries to do events
        -----
        OUTPUTS
        ts: TimeSeriesX Object, a functional timeseries object with indexable events
        """

        # Check if events is in the dims
        if all(['events' in y for y in [x.dims for x in ts_list]]):
            # Extract events before overwriting them
            evs = np.concatenate([x.events.data for x in ts_list]).view(np.recarray)
            ts = xr.concat(objs=ts_list, dim=dim, *args, **kwargs)
            ts['events'] = evs  # Reset the events to the correct np.recarray
            return ts

        # TODO: Change to check all dims that are record arrays and do it

        # Try to just concat normally, if that doesn't work show them why.
        else:
            try:
                ts = xr.concat(objs=ts_list, dim=dim, *args, **kwargs)
                return ts
            except:
                print(ts_list[0].dims)
                print('There needs to be an "events" dim for each of the TimeSeriesX in the passed list')
                assert all(['events' in y for y in [x.dims for x in ts_list]])

    def to_edf(self, save_path, annotations=None, header=None):
        valid = np.array(['channel', 'channels', 'bipolar_pair', 'bipolar_pairs'])
        # Find out what the potential dimension name for the contacts is
        label = valid[np.where(np.in1d(valid, np.intersect1d(np.array(self.dims), valid)))][0]
        try:
            assert label == self.dims[0]
        except AssertionError:
            raise AssertionError('Please ensure that the first dimension of the timeseries is {}'.format(label))

        n_valid_signals = self[label].shape[0]

        writer = EdfWriter(save_path, n_channels=n_valid_signals,
                           file_type=pyedflib.FILETYPE_EDFPLUS)

        channel_info, data_list = [], []
        physical_maximums = self.max('time')
        physical_minimums = self.min('time')

        for i in np.arange(n_valid_signals):
            ch_dict = {'label': str(self[label][i].data),
                       'dimension': 'uV',
                       'sample_rate': float(self.samplerate.data),
                       'physical_max': float(physical_maximums[i]),
                       'physical_min': float(physical_minimums[i]),
                       'digital_max': 32767.,
                       'digital_min': -32768.,
                       'transducer': '',
                       'prefilter': ''}

            channel_info.append(ch_dict)
            data_list.append(self[i].data)

        # Copy old file's annotation
        if annotations is not None:
            for annotation in annotations:
                writer.writeAnnotation(annotation[0], -1, annotation[2])

        # Create a default header if none is passed
        if header is None:
            header = {'technician': '',
                      'recording_additional': '',
                      'patientname': 'No Name',
                      'patient_additional': '',
                      'patientcode': '0',
                      'equipment': 'NKC-EEG-1200A V01.00',
                      'admincode': '',
                      'gender': '',
                      'startdate': datetime(2017, 8, 17, 1, 4, 43),
                      'birthdate': date(1951, 8, 2)}

        header = {'technician': 'tec1', 'recording_additional': 'recAdd1', 'patientname': 'pat1',
                  'patient_additional': 'patAdd1', 'patientcode': 'code1', 'equipment': 'eq1',
                  'admincode': 'admin1', 'gender': 1, 'startdate': datetime(2017, 1, 1, 1, 1, 1),
                  'birthdate': date(1951, 8, 2)}
        writer.setHeader(header)
        writer.setSignalHeaders(channel_info)
        #writer.writeSamples(data_list)
        writer.writePhysicalSamples(data_list)

        # Close writer/reader
        writer.close()

    @staticmethod
    def get_eeg_ch_matches(signal_labels, contacts_label):
        """ This is used in cases where we want to load data raw off the clinical record and replace bad names with good names
        f = EdfReader(file_name=file_path)
        signal_labels = f.getSignalLabels()
        reader = CMLReader(subject='R1207J', experiment="FR1", session=1,
                           localization=0, montage=0)
        df = reader.load('contacts')
        contacts_label = np.array(df['label'])
        """
        # Find partial matches in edf raw signals with cmlreaders channel labels
        signal_labels = np.array([label.upper() for label in signal_labels])
        match_index = []
        for channel in contacts_label:
            match = np.flatnonzero(np.core.defchararray.find(signal_labels, channel) != -1)
            if match.size > 0:
                match_index.append(match)
        match_index = np.unique(np.concatenate(match_index))
        return match_index

    # TODO: Make it automatic
    def fix_ch_names(self, verbose=False):
        valid = np.array(['channel', 'channels', 'bipolar_pair', 'bipolar_pairs'])
        # Find out what the potential dimension name for the contacts is
        label = valid[np.where(np.in1d(valid, np.intersect1d(np.array(self.dims), valid)))][0]
        signal_labels = self[label].data
        format_strings = np.array(list(map(lambda s: s.split(' ')[-1].replace('-Ref', '').upper(),
                                           np.array(signal_labels))))
        if verbose:
            print('Replacing Old Channel labels with new channel labels')
            display(list(zip(self[label].data, format_strings)))

        self[label].data = format_strings
        return

    def filter_with(self, filter_class, **kwargs):
        """Filter the time series data using the specified filter class.
        Parameters
        ----------
        filter_class : type
            The filter class to use.
        kwargs
            Keyword arguments to pass along to ``filter_class``.
        Returns
        -------
        filtered : TimeSeries
            The resulting data from the filter.
        Raises
        ------
        TypeError
            When ``filter_class`` is not a valid filter class.
        """
        from ptsa.data.filters.base import BaseFilter

        if not issubclass(filter_class, BaseFilter):
            raise TypeError("filter_class must be a child of BaseFilter")

        filtered = filter_class(self, **kwargs).filter()
        return filtered

    def test(self):
        print(self.data)
        print(self.coords)

    # TODO: Add in to_mne helper utility built in
    def to_mne(self, timeseries, indicator=None):
        from mne import create_info, EpochsArray
        """If indicator is passed it must be a 1d array of shape timeseries.events, and consist of
        only 1s and 0s whereby 1 is an event and 0 is not"""
        # Create MNE Dataset
        ch_names = list(self.coords['channels'].data)
        info = create_info(ch_names, self.coords['samplerate'].data, ch_types='eeg')

        # Reorganize data for MNE format
        from ptsa.data.common import get_axis_index

        data = timeseries.transpose('events', 'channels', 'time')

        # Create events array for MNE
        mne_evs = np.empty([data['events'].shape[0], 3]).astype(int)
        mne_evs[:, 0] = np.arange(data['events'].shape[0])
        mne_evs[:, 1] = data['time'].shape[0]
        mne_evs[:, 2] = 1
        event_id = dict(event=1)

        if indicator is not None:
            mne_evs[:, 2] = indicator
            event_id = dict(event=1, not_event=0)

        tmin = self.coords['time'].data[0]
        arr = EpochsArray(np.array(data), info, mne_evs, tmin, event_id)
        arr.set_eeg_reference(ref_channels=None)  # Set to average reference
        arr.apply_proj()

        return arr

if __name__ == '__main__':
    ts = TimeSeriesLF.from_edf('/Users/loganfickling/Clumsy/test_generator.edf')
    ts.test()


import numpy as np
from math import pi


def degree_to_rad(angle_degree):
    #from math import pi
    return (angle_degree * pi) / 180.


def rad_to_degree(angle_rad):
    #from math import pi
    return (angle_rad * 180.) / pi


def get_close_degrees(arr, index, degree_dist=20):
    """
    ------
    INPUTS
    arr: np.array, values of degrees between
    index: int, the value of the array to index at
    degree_dist: int, value of how many degrees there can be between
                 the points to count.

    """
    #
    boolean = [np.isclose(arr[index], arr, .1, degree_dist)]
    return arr[boolean], boolean


def count_45s(arr):
    """
    Go through every 45 degrees of a circle, and count the number
    of events in the array that fall within 20 degrees of the arr
    -----
    INPUTS:
    arr: np.array, an array of your degrees to check over
    -----
    outputs:
    dd: dict, keys are the degrees (e.g. 45), the values are the #
    """
    d = {}
    # Get every 45th degree in 360
    points = np.arange(-360, 360 + 45, 45)
    # Check if it's within 20 degrees, add to d the number of them
    for degree in points:
        d[degree] = arr[np.isclose(degree, arr, .1, 20)].shape[0]

    for k, v in enumerate(sorted(d.keys())):
        convert = 360 + v
        d[convert] = d[convert] + d[v]
        if v == -45:
            break
    dd = {v: d[v] for k, v in enumerate(d) if v >= 0}
    test = zip(degree_to_rad(np.array(dd.keys())), dd.values())
    return dd, test


def polar_chart(ts, electrode):
    """
    ts: timeseries
    electrode: int
    """

    a, b = count_45s(rad_to_degree(ts[electrode, 0, :, 5].data))
    """
    =======================
    Pie chart on polar axis
    =======================

    Demo of bar plot on a polar axis.
    """
    try:
        import matplotlib.pyplot as plt
    except Exception as e:
        print(e)
        return

    # Compute pie slices
    t = [0.0, 3.9269908169872414, 2.3561944901923448, 6.2831853071795862, 0.78539816339744828,
         4.7123889803846897, 3.1415926535897931, 1.5707963267948966, 5.497787143782138]
    N = len(t)
    # theta = np.linspace(0.0, 2 * np.pi, N, endpoint=False)
    counts = [bb[1] for bb in b]
    radii = np.array(counts)
    # width = np.ones(N)/7

    ax = plt.subplot(111, projection='polar')

    bars = ax.bar(counts, t)
    # bars = ax.bar(t, radii, width=width,)

    # Use custom colors and opacity
    # for r, bar in zip(radii, bars):
    # bar.set_facecolor(plt.cm.jet(r / 10.))
    # bar.set_alpha(0.5)

    plt.show()


import numpy as np
from xarray import DataArray
from numpy import ceil, log2
from ptsa.data.timeseries import TimeSeries
from ptsa.data.common import get_axis_index
from ptsa.data.filters import BaseFilter, ButterworthFilter
from scipy.signal import hilbert
from copy import deepcopy
import xarray as xr

__all__ = ['HilbertFilter']


class HilbertFilter(BaseFilter):
    """Applies Hilbert filter to a time series.

    Keyword Arguments
    -----------------

    timeseries
         TimeSeries object
    """

    def __init__(self, timeseries):
        """

        Parameters
        ----------
        timeseries: TimeSeries like object, pre-filtered in frequency-band of interest
        """
        super(HilbertFilter, self).__init__(timeseries)

    def filter(self):
        """

        Returns
        -------

        """
        time_axis = get_axis_index(self.timeseries, 'time')
        # Zero pad to speed up calculation
        padding = int(2 ** ceil(log2(self.timeseries.shape[time_axis])))
        output = hilbert(self.timeseries, N=padding, axis=time_axis)
        # Find out where we need to remove the padding since scipy doesn't do it by default
        valid_pad = (Ellipsis, slice(self.timeseries.shape[time_axis]))
        returned_signal = output[valid_pad]
        # Create a time-series for the user
        coords_dict = {coord_name: DataArray(coord.copy()) for coord_name, coord in
                       list(self.timeseries.coords.items())}
        coords_dict['samplerate'] = self.timeseries['samplerate']
        dims = [dim_name for dim_name in self.timeseries.dims]
        filtered_timeseries = TimeSeries(returned_signal, dims=dims, coords=coords_dict)
        filtered_timeseries.attrs = self.timeseries.attrs.copy()
        return filtered_timeseries


from ptsa.data.timeseries import TimeSeries
from scipy.signal import hilbert

def Hilbert(signal):
    """Applies a hilbert transform to the signal (must be a single channel)

    ### IMPORTANT NOTES####
    This function adds zero padding to speed up the processing of the fast
    fourier transformation(FFT) if the length of the signal passed is
    not a power of two (for example a 49999 lengthed signal
    will take orders of magnitude longer than a 50000 lengthed signal to
    compute a Fast fourier transformation on )

    ------
    INPUTS:
    signal: array like or TimeSeries, MUST BE A SINGLE CHANNEL.
    ------
    OUTPUTS:
        signal: TimeSeriesX, a TimeSeries object of shape signal that is hilbert filtered
    """
    padding = np.zeros(int(2 ** np.ceil(np.log2(len(signal)))) - len(signal))
    tohilbert = np.hstack((signal, padding))
    result = hilbert(tohilbert)
    result = result[0:len(signal)]

    if type(signal) == TimeSeries:
        # Create a copy of timeseries
        return TimeSeries(data=result, dims=signal.dims, coords=signal.coords)

    return result


def get_amplitude_envelope(signal):
    """Returns that instantaneous amplitude envelope of the analytic signal
       from the Hilbert transformation
    ------
    INPUTS:
    signal: array like or TimeSeriesX.
    """
    _xarray = np.abs(Hilbert(signal))
    return TimeSeries(_xarray, coords=_xarray.coords)



# TODO: Imports
def _calc_start_stop_from_offset(offsets, start_time=-1., end_time=2., buffer_time=.5):
    starts = (offsets + ((buffer_time * -1000) + (start_time * 1000))).astype(int)
    stops = (offsets + ((buffer_time * 1000) + (end_time * 1000))).astype(int)
    return starts, stops


def chop_data_from_offsets(timeseries, offsets, start_time, end_time, buffer_time=0.):
    # TODO: Check if time is positive and if it is make it negative so they don't have an empty array returned
    if np.sign(start_time) == np.sign(end_time):
        start_time = -1 * start_time

    start, stop = _calc_start_stop_from_offset(offsets, start_time=start_time,
                                               end_time=end_time, buffer_time=buffer_time)

    time_axis = np.arange(stop[0] - start[0]) / timeseries['samplerate'].data
    # Check if it's symmerical start stop and change it so midpoint in time is 0
    # TODO: Should this be an argument?
    if start_time * -1 == end_time:
        time_axis = time_axis - (time_axis[-1] / 2)
    dat = []
    for _start, _stop in zip(start[:-1], stop[:-1]):
        chopped = timeseries[Ellipsis, slice(_start, _stop)]
        chopped['time'].data = time_axis
        dat.append(chopped)
    data = TimeSeriesLF.concat(dat, 'events')
    data['buffer_time'] = buffer_time
    return data

# TODO: Put these in a better place
def minimum_time_from_frequency_ncycles(frequency, cycles=5):
    return cycles / frequency


def numbers_between(arr, low, high, values=False):
    if values:
        return arr[numbers_between(arr, low, high, False)]
    return np.logical_and(low <= arr, arr <= high)


def dataframe_from_connectivity_matrix(connectivity_matrix, frequency_bands, frequency_selection):
    dat = conectivity_matrix[:, :, numbers_between(frequency_bands, *frequency_selection)]
    return pd.DataFrame(np.mean(dat, -1))


def group_dataframe_from_rois(df, rois):
    """Given a dataframe N x N, returns mean M x M matrix of grouped mean rois

    INPUTS
    -------
    df: pd.DataFrame, shape = N x N
    rois: np.array like, shape N"""
    rois = pd.Series(rois)
    rois_df = df.groupby(by=rois, axis=0).mean().groupby(by=rois, axis=1).mean()
    return rois_df


"""Contains utility functions used in analysis of sleep data"""

from copy import deepcopy

import numpy as np
from numpy import (absolute, arange, around, asarray,
                   exp, empty, hstack, mean, ones, pi, zeros)
from scipy.fftpack import next_fast_len
from scipy.ndimage.filters import gaussian_filter, gaussian_filter1d
from scipy.signal import (butter, cheby2, filtfilt,
                          fftconvolve, hilbert, remez,
                          sosfiltfilt,  # tukey,
                          )

from Clumsy.signal import HilbertFilter
from Clumsy.signal import check_stability
# from ptsa.data.timeseries import TimeSeries
from Clumsy.signal.timeseriesLF import TimeSeries

__all__ = ['transform_signal']
def transform_signal(dat, method, s_freq=None, method_opt=None, dat2=None, axis=-1):
    """Transform the data using different methods.

    Wrapper around wonambi for use with timeseries, changes in code to vectorize wherever
    possible, and extend 1d functionality to nD
    See: https://wonambi-python.github.io/


    Parameters
    ----------
    dat : TimeSeriesLF or ndarray (dtype='float')
        vector with all the data for one channel
    method : str
        one of 'butter', 'cheby2', 'double_butter', 'morlet', 'morlet_real',
        'hilbert', 'abs', 'smooth', 'gaussian', 'remez', 'wavelet_real',
        'low_butter', 'zscore', 'moving_rms', 'moving_ms'
    s_freq : float defaults as None
        sampling frequency, not necessary if a TimeSeriesLF object is passed
    method_opt : dict
        depends on methods
    dat2 : ndarray(dtype='float')
        second vector with data
    axis : int, defaults as -1
        axis on which to transform the data

    Returns
    -------
    ndarray (dtype='float')
        vector with all the data for one channel
    or
    TimeSeriesLF (dtype='float')
        array with all the data for all channels

    Notes
    -----
    double_butter implements an effective bandpass by applying a highpass,
    followed by a lowpass. This method reduces filter instability, due to
    underlying numerical instability arising from nyquist / freq, at low freq.
    Wavelets pass only absolute values already, it does not make sense to store
    the complex values.

    When using a butterworth filter, it's important to consider stability, if a
    filter is unstable it will explode. To check one can use:
    z, p, k = signal.tf2zpk(b, a)
    assert(np.max(np.abs(p))<=1)
    if the filter is unstable, one can get around this by using a second order
    section (e.g. sosbutter, double_sosbutter)
    see: https://stackoverflow.com/questions/41371915/20hz-20000hz-butterworth-filtering-exploding


    filtfilt is zero-phase filtering, which doesn't shift the signal as it filters. Since the phase is zero at all frequencies, it is also linear-phase.
    Filtering backwards in time requires you to predict the future, so it can't be used in "online" real-life applications, only for offline processing
    of recordings of signal.
    e.g.
    # Applies filter forward and backward in time
    imp_ff = signal.filtfilt(b, a, impulse)

    lfilter is causal forward-in-time filtering only, similar to a real-life electronic filter. It can't be zero-phase. It can be linear-phase
    (symmetrical FIR), but usually isn't. Usually it adds different amounts of delay at different frequencies.
    # Applies filter forward in time twice (for same frequency response)
    imp_lf = signal.lfilter(b, a, signal.lfilter(b, a, impulse))




    Methods
    -------


    butter has parameters: zero-phase filtering
        freq : tuple of float
            low and high values for bandpass
        order : int
            filter order (will be effecively doubled by filtfilt)

    cdemod has parameters:
        freq : float
            carrier frequency for complex demodulation

    cheby2 has parameters:
        freq : tuple of float
            low and high values for bandpass
        order : int
            filter order (will be effecively doubled by filtfilt)

    double_butter has parameters:
        freq : tuple of float
            low and high values for highpass, then lowpass
        order : int
            filter order (will be effecively doubled by filtfilt)

    double_sosbutter has parameters:
        freq : tuple of float
            low and high values for highpass, then lowpass
        order : int
            filter order (will be effecively doubled by filtfilt)

    gaussian has parameters:
        dur : float
            standard deviation of the Gaussian kernel, aka sigma (sec)

    low_butter has parameters:
        freq : float
            Highcut frequency, in Hz
        order : int
            filter order (will be effecively doubled by filtfilt)

    morlet has parameters:
        f0 : float
            center frequency in Hz
        sd : float
            standard deviation of frequency
        dur : float
            window length in number of standard deviations

    wavelet_real has parameters:
        freqs : ndarray
            vector of wavelet frequencies for spindle detection
        dur : float
            duration of the wavelet (sec)
        width : float
            wavelet width
        win : float
            moving average window length (sec) of wavelet convolution

    smooth has parameters:
    NOT IMPLEMENTED YET
        dur : float
            duration of the window (sec)
        win : str
            'flat' or 'triangle' based upon smoothing desired

    moving_rms has parameters:
        dur : float
            duration of the window (sec)

    moving_zscore has parameters:
        dur : float
            duration of the z-score sliding window (sec)

    remez has parameters:
        freq : tuple of float
            low and high values for bandpass
        rolloff : float
            bandwidth, in hertz, between stop and pass frequencies
        dur : float
            dur * s_freq = N, where N is the filter order, a.k.a number of taps

    sosbutter has parameters:
        freq : tuple of float
            low and high values for highpass, then lowpass
        order : int
            filter order (will be effecively doubled by filtfilt)

    tri_smooth has parameters:
        dur : float
            length of convolution window, base of isosceles triangle
    """
    valid_methods = (
    'abs', 'abs2', 'cheby2', 'butter', 'high_butter', 'low_butter', 'sosbutter', 'double_butter', 'double_sosbutter',
    'gaussian', 'hilbert', 'remez', 'moving_covar', 'moving_power_ratio', 'moving_sd', 'moving_zscore', 'moving_rms',
    'moving_ms', 'cdemod', 'morlet', 'smooth', 'wavelet_real')
    not_implemented_yet = ('moving_covar', 'moving_power_ratio', 'moving_sd', 'moving_zscore', 'moving_rms',
                           'moving_ms', 'cdemod', 'morlet', 'smooth', 'wavelet_real')

    is_timeseries = True if issubclass(type(dat), TimeSeries) else False
    if is_timeseries:
        copy = deepcopy(dat)
        if method.lower() != 'hilbert':  # HilbertFilter behavior will fail if dat is an array
            dat = np.array(dat)
        if s_freq is None:
            s_freq = float(copy.samplerate)

    nyquist = s_freq / 2

    # ------> Absolute Value
    if 'abs' == method:
        dat = absolute(dat)

    if 'abs2' == method:
        dat = dat.real ** 2 + dat.imag ** 2

    # ------> Chebyshev type II digital filter
    # TODO: Add in cheby1
    if 'cheby2' == method:
        freq = method_opt['freq']
        N = method_opt['order']
        try:
            Rs = method_opt['Rs']
        except:
            Rs = 40
        Wn = asarray(freq) / nyquist
        b, a = cheby2(N, Rs, Wn, btype='bandpass')
        dat = filtfilt(b, a, dat, axis=axis)

    # ------> Butterworth filter
    if 'butter' in method:
        freq = method_opt['freq']
        N = method_opt['order']
        if method not in ('low_butter', 'high_butter'):
            Wn = asarray(freq) / nyquist
        else:
            Wn = freq / nyquist
        output = 'sos' if 'sos' in method else 'ba'
        correctfiltfilt = sosfiltfilt if 'sos' in method else filtfilt

        def apply_correctfiltfilt(filter_inputs, correctfiltfilt, dat, axis):
            """applies correct filtfilter for sos/ba and unpacks tuple where appropriate"""
            if type(filter_inputs) == tuple:  # * unpacks the tuple
                return correctfiltfilt(*filter_inputs, dat, axis=axis)
            else:
                return correctfiltfilt(filter_inputs, dat, axis=axis)

        # Not sure which line reads better -> Double butter filters
        if method in ('double_butter', 'double_sosbutter'):
            # High pass
            filter_inputs = butter(N, Wn[0], btype='highpass', output=output)
            if output == 'ba':  # Check stability of filter and prompt user to use sos if unstable
                check_stability(b=filter_inputs[0], a=filter_inputs[1])
            dat = apply_correctfiltfilt(filter_inputs, correctfiltfilt, dat, axis)
            # Low pass
            filter_inputs = butter(N, Wn[1], btype='lowpass', output=output)
            if output == 'ba':  # Check stability of filter and prompt user to use sos if unstable
                check_stability(b=filter_inputs[0], a=filter_inputs[1])
            dat = apply_correctfiltfilt(filter_inputs, correctfiltfilt, dat, axis)

        elif method in ('butter', 'high_butter', 'low_butter', 'sosbutter'):

            if 'low' in method:
                btype = 'lowpass'
            elif 'high' in method:
                btype = 'highpass'
            elif method in ('butter', 'sosbutter'):
                btype = 'bandpass'

            filter_inputs = butter(N, Wn, btype=btype, output=output)
            if output == 'ba':  # Check stability of filter and prompt user to use sos if unstable
                check_stability(b=filter_inputs[0], a=filter_inputs[1])
            dat = apply_correctfiltfilt(filter_inputs, correctfiltfilt, dat, axis)

    if 'gaussian' == method:
        truncate = 8
        sigma = method_opt['dur']  # 0.004
        if issubclass(type(dat), TimeSeries):
            dat = gaussian_filter1d(dat, sigma * s_freq,
                                    truncate=truncate, axis=axis, mode='constant')
        else:
            dat = gaussian_filter(dat, sigma)

    if 'hilbert' == method:
        if issubclass(type(dat), TimeSeries):
            hf = HilbertFilter(timeseries=dat)
            return hf.filter()

        else:  # ASSUMES 1D???
            N = len(dat)
            dat = hilbert(dat, N=next_fast_len(N))  # much faster this way
            dat = dat[:N]  # truncate away zero-padding

    if 'remez' == method:
        Fp1, Fp2 = method_opt['freq']
        rolloff = method_opt['rolloff']
        dur = method_opt['dur']

        N = int(s_freq * dur)
        Fs1, Fs2 = Fp1 - rolloff, Fp2 + rolloff
        dens = 20

        bpass = remez(N, [0, Fs1, Fp1, Fp2, Fs2, nyquist], [0, 1, 0],
                      grid_density=dens, fs=s_freq)
        dat = filtfilt(bpass, 1, dat, axis=axis)

    if 'moving' in method:  # This is a stupid implementation change it to rolling array instead of for loops#
        raise NotImplementedError('{} has not been implemented yet'.format(method))
        return

        dur = method_opt['dur']
        step = method_opt['step']
        n_step = int(around(step * s_freq)) if step else None
        n_halfdur = int(s_freq * dur / 2)
        len_dat = len(dat)
        len_out = int(len_dat / n_step) if n_step else len_dat
        last = len_dat - 1
        out = zeros((len_out))

        if 'moving_covar' == method:  #### NEEDS UPDATE ####
            raise NotImplementedError('{} has not been implemented yet'.format(method))

        if 'moving_power_ratio' == method:  #### NEEDS UPDATE ####
            raise NotImplementedError('{} has not been implemented yet'.format(method))
            # freq1 = method_opt['freq_narrow']
            # freq2 = method_opt['freq_broad']
            # fft_dur = method_opt['fft_dur']

        if 'moving_sd' == method:  #### NEEDS UPDATE ####
            raise NotImplementedError('{} has not been implemented yet'.format(method))

        if 'moving_zscore' == method:  #### NEEDS UPDATE ####
            raise NotImplementedError('{} has not been implemented yet'.format(method))
            #pcl_range = method_opt['pcl_range']

        if method in ['moving_rms', 'moving_ms']:  #### NEEDS UPDATE ####
            raise NotImplementedError('{} has not been implemented yet'.format(method))
            #if method == 'moving_rms':
                #pass

    if 'cdemod' == method:  ### Fack this needs to be fixed#### Apply to nD
        carr_freq = method_opt['freq']
        carr_sig = exp(-1j * 2 * pi * carr_freq * arange(0, len(dat)) / s_freq)

        dat = dat * carr_sig

    if 'morlet' == method:  #### NEEDS UPDATE ####
        raise NotImplementedError('{} has not been implemented yet consider using ptsa instead'.format(method))
        #f0 = method_opt['f0']
        #sd = method_opt['sd']
        #dur = method_opt['dur']

        #wm = _wmorlet(f0, sd, s_freq, dur)
        #dat = absolute(fftconvolve(dat, wm, mode='same'))

    if 'smooth' == method:  #### NEEDS UPDATE ####
        raise NotImplementedError('{} has not been implemented yet'.format(method))
    """

    dur = method_opt['dur']
    win = method_opt['win']

    if 'flat' == win:
        flat = ones(int(dur * s_freq))
        H = flat / sum(flat)
    elif 'triangle' == win:
        T = int(dur * s_freq / 2)
        a = arange(T, 0, -1)

        H = hstack([a[-1:0:-1], a])
        H = H / sum(H)

    dat = fftconvolve(dat, H, mode='same')
    """
    if 'wavelet_real' == method:  #### NEEDS UPDATE ####
        raise NotImplementedError('{} has not been implemented yet consider using ptsa instead'.format(method))
    """
    freqs = method_opt['freqs']
    dur = method_opt['dur']
    width = method_opt['width']
    win = int(method_opt['win'] * s_freq)

    wm = _realwavelets(s_freq, freqs, dur, width)
    tfr = empty((dat.shape[0], wm.shape[0]))
    for i, one_wm in enumerate(wm):
        x = abs(fftconvolve(dat, one_wm, mode='same'))
        tfr[:, i] = fftconvolve(x, tukey(win), mode='same')
    dat = mean(tfr, axis=1)
    """
    if is_timeseries:
        copy.data = dat
        return copy

    return dat